from . import Check, Finding
from .utils import is_index, get_module, get_first_heading


class TopLevelCasingFinding(Finding):
    NAME = "TopLevelCasing"

    def msg(self):
        return "Module name has wrong case"


class TopLevelCasingCheck(Check):
    FINDING = TopLevelCasingFinding

    def run(self, file, lines):
        if not is_index(file):
            return
        module = get_module(file)
        top_level_heading, i = get_first_heading(lines)
        if top_level_heading and top_level_heading.lower().startswith(module.lower()):
            if not top_level_heading.startswith(module):
                self.add_finding(file, i + 1)



from . import Check, Finding
from .utils import is_index, get_first_heading


class TopLevelLengthFinding(Finding):
    NAME = "TopLevelLength"

    def __init__(self, file, line, exp_length, act_length):
        super().__init__(file, line)
        self._exp_length = exp_length
        self._act_length = act_length

    def msg(self):
        return f"Title too long: {self._act_length} > {self._exp_length}"


class TopLevelLengthCheck(Check):
    FINDING = TopLevelLengthFinding

    CONFIG = {
        "limit": 40,
    }

    def run(self, file, lines):
        if not is_index(file):
            return
        top_level_heading, i = get_first_heading(lines)
        if len(top_level_heading) > self.CONFIG["limit"]:
            self.add_finding(
                file,
                i + 1,
                exp_length=self.CONFIG["limit"],
                act_length=len(top_level_heading),
            )



from . import Check, Finding


class DummyFinding(Finding):
    NAME = "DummyFinding"

    def msg(self):
        return "This is some dummy finding message"


class DummyCheck(Check):
    FINDING = DummyFinding

    CONFIG = {
        "some": "config option",
    }

    def run(self, file, lines):
        # file:  path to the document to check
        # lines: content of that document read as list of lines

        # violation found in line 42, add a corresponding finding
        self.add_finding(file, 42)



from . import Finding


class Check:
    FINDING = Finding

    CONFIG = {}

    def __init__(self):
        self._findings = []

    @property
    def findings(self):
        return self._findings

    def run(self, file, lines):
        raise NotImplementedError

    def add_finding(self, file, line, **kwargs):
        self._findings.append(self.FINDING(file, line, **kwargs))



import os
import warnings
import importlib.util

from pathlib import Path
from datetime import datetime
import yaml
from yaml.loader import SafeLoader

from .label_check import *

logger = logging.getLogger(__name__)


def config_inited(app, config):
    # suppress deprecated warnings in some Sphinx extensions
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="sphinxcontrib")

    # page footer
    if config.dox_style_footer:
        if os.path.isfile(config.dox_style_footer):
            with open(config.dox_style_footer, "r") as f:
                config["footer_string"] = yaml.load(f, Loader=SafeLoader)

    # HTML theme and options
    config.html_theme = "sphinx_rtd_theme"
    config.html_show_sphinx = False
    config.html_show_sourcelink = True
    config.html_style = "rtd_theme_overrides.css"
    config.html_theme_options.update(
        {
            "style_external_links": True,
            "sticky_navigation": False,
            "prev_next_buttons_location": None,
            "navigation_depth": 6,
        }
    )

    # HTML additional files
    static = Path.joinpath(Path(__file__).parent, "_static").as_posix()
    config.html_static_path = [static]
    config.templates_path = [static]
    app.add_css_file("colors.css")
    app.add_css_file("colored_table.css")
    app.add_css_file("header.css")

    # data classification
    if "data_classification_default" not in config.html_context:
        config.html_context["data_classification_default"] = "Confidential"

    # latexpdf needs to know who certain unicode characters shall be rendered
    if not config.latex_elements:
        config.latex_elements = {}
    if not "preamble" in config.latex_elements:
        config.latex_elements["preamble"] = ""
    config.latex_elements[
        "preamble"
    ] += r"""
        \DeclareUnicodeCharacter{25BA}{$\blacktriangleright$}
        \DeclareUnicodeCharacter{25BC}{$\blacktriangledown$}
        \DeclareUnicodeCharacter{25B2}{$\blacktriangle$}
        \DeclareUnicodeCharacter{25C4}{$\blacktriangleleft$}
        """
    if not "extrapackages" in config.latex_elements:
        config.latex_elements["extrapackages"] = ""
    config.latex_elements[
        "extrapackages"
    ] += r"""
        \usepackage{pmboxdraw}
        """

    # overwrite PlantUML path, needed for Windows
    if "PLANTUML_DIR" in os.environ:
        config.plantuml = "java -Xlog:disable -jar {}".format(
            Path(os.getenv("PLANTUML_DIR")).joinpath("plantuml.jar").resolve()
        )

    # global definitions (important: do not use leading spaces)
    if not config.rst_prolog:
        config.rst_prolog = ""
    config.rst_prolog += """
.. role:: green
.. role:: yellow
.. role:: orange
.. role:: red
.. role:: blue
.. role:: white
.. role:: black
.. role:: grey
.. role:: greyitalic
.. role:: raw-html(raw)
    :format: html
.. |br| replace:: :raw-html:`</br>`
"""

    # case-insensitive glossary
    if "ref.term" not in config.suppress_warnings:
        config.suppress_warnings.append("ref.term")

    ################

    # At the end try to load a custom specific configuration file
    # which can be used to overwrite all values set by this extension.
    try:
        extend_filename = Path(app.confdir) / "extend_conf.py"
        if os.path.exists(extend_filename):
            spec = importlib.util.spec_from_file_location("extend_conf", extend_filename)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            module.extend_conf(config)
            print("Custom configuration file loaded")
    except (ImportError, TypeError):
        logger.warning("extend_conf.py requires a function\n    def extend_conf(config: dict)")


def get_footer_string(app, docname):
    docpath = docname
    while True:
        if docpath in app.config["footer_string"]:
            return app.config["footer_string"][docpath]
        else:
            pos = docpath.rfind("/")
            if pos >= 0:
                docpath = docpath[:pos]
            elif docpath != ".":
                docpath = "."
            else:
                return None


def add_extrafooter(app, _doctree, docname):
    build_type = "official build" if "OFFICIAL_BUILD" in os.environ else "UNOFFICIAL BUILD"
    build_time = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
    app.env.metadata[docname]["footer_string"] = (
        "Documentation from " + build_time + ", " + build_type + "."
    )

    if hasattr(app.config, "footer_string"):
        ftext = get_footer_string(app, docname)
        if ftext:
            app.env.metadata[docname]["footer_string"] += "<br>" + ftext


def config_setup(app):
    # config_inited must be called before any other extension, priority 0 is the highest priority
    app.connect("config-inited", config_inited, priority=0)

    app.connect("doctree-resolved", add_extrafooter)

    app.connect("env-before-read-docs", prepare_labels)
    app.connect("env-purge-doc", purge_labels)
    app.connect("doctree-read", collect_labels)
    app.connect("env-merge-info", merge_labels_parallel_build)

    app.add_config_value("dox_style_footer", None, "html")



from docutils import nodes
from sphinx.util import logging

logger = logging.getLogger(__name__)


def add_refid(std, docname, location, ref_id):
    std["label_location"][ref_id] = location
    if docname not in std["labels_in_document"]:
        std["labels_in_document"][docname] = {ref_id}
    else:
        std["labels_in_document"][docname].add(ref_id)


def collect_labels(app, doctree):
    std = app.env.domaindata["std"]
    docname = app.env.docname
    for node in doctree.traverse(nodes.target):
        ref_id = node.get("refid")
        # ignore duplicated labels for a single env, this will be checked by Sphinx internally
        if ref_id:
            location = docname + ".rst"
            if ref_id in std["label_location"]:
                location_other = std["label_location"][ref_id]
                if location_other != location:
                    logger.warning(f"label {ref_id} defined in {location} and {location_other}")
            else:
                add_refid(std, docname, location, ref_id)


def purge_labels(_app, env, docname):
    std = env.domaindata["std"]
    if "labels_in_document" in std:
        if docname in std["labels_in_document"] and "label_location" in std:
            for ref_id in std["labels_in_document"][docname]:
                std["label_location"].pop(ref_id, None)
        std["labels_in_document"].pop(docname, None)


def prepare_labels(app, _env, _docnames):
    std = app.env.domaindata["std"]
    if "labels_in_document" not in std:
        std["labels_in_document"] = {}
    if "label_location" not in std:
        std["label_location"] = {}


def merge_labels_parallel_build(_app, env, docnames, other):
    std = env.domaindata["std"]
    std_other = other.domaindata["std"]

    for docname in docnames:
        if docname not in std_other["labels_in_document"]:
            # no labels to merge
            continue
        for ref_id in std_other["labels_in_document"][docname]:
            location_other = std_other["label_location"][ref_id]
            if ref_id in std["label_location"]:
                location = std["label_location"][ref_id]
                if location_other != location:
                    logger.warning(f"label {ref_id} defined in {location} and {location_other}")
                # else: label already there, can happen when rebuilding the documentation
            else:
                add_refid(std, docname, location_other, ref_id)



from . import Check, Finding
from .utils import is_index, get_module, get_first_heading


class TopLevelModuleNameFinding(Finding):
    NAME = "TopLevelModulName"

    def __init__(self, file, line, module_name):
        super().__init__(file, line)
        self.module_name = module_name

    def msg(self):
        return f'Module name "{self.module_name}" not found'


class TopLevelModuleNameCheck(Check):
    FINDING = TopLevelModuleNameFinding

    def run(self, file, lines):
        if not is_index(file):
            return
        module = get_module(file)
        top_level_heading, i = get_first_heading(lines)
        if top_level_heading and not top_level_heading.lower().startswith(module.lower()):
            self.add_finding(file, i + 1, module_name=module)



from . import Check, Finding
from .utils import is_heading


class HeadingLevelsFinding(Finding):
    NAME = "HeadingLevels"

    def __init__(self, file, line, expected_char, actual_char):
        super().__init__(file, line)
        self.expected_char = expected_char
        self.actual_char = actual_char

    def msg(self):
        return f"Heading must be {self.expected_char} instead of {self.actual_char}"


class HeadingLevelsCheck(Check):
    FINDING = HeadingLevelsFinding

    CONFIG = {"characters": ["=", "-", "+", "~", "^", '"']}

    def __init__(self):
        self.heading_characters = self.CONFIG["characters"]
        super().__init__()
        self.heading_levels = {c: i + 1 for i, c in enumerate(self.heading_characters)}

    def run(self, file, lines):
        current_level = 0
        for i, line in enumerate(lines):
            if is_heading(line):
                if line[0] in self.heading_characters:
                    level = self.heading_levels[line[0]]
                    if level - current_level > 1:
                        self.add_finding(
                            file,
                            i + 1,
                            expected_char=self.heading_characters[current_level],
                            actual_char=line[0],
                        )
                    current_level = level
                else:
                    self.add_finding(
                        file,
                        i + 1,
                        expected_char=self.heading_characters[current_level],
                        actual_char=line[0],
                    )



from . import Check, Finding
import re


class IncludeRstFinding(Finding):
    NAME = "IncludeRst"

    def msg(self):
        return "Include-directive must not point to an *.rst file"


class IncludeRstCheck(Check):
    FINDING = IncludeRstFinding

    include_rst_regex = re.compile(r"\.\.\s+include\s*::.*\.rst$")

    def run(self, file, lines):
        for i, line in enumerate(lines):
            if self.include_rst_regex.match(line):
                self.add_finding(file, i + 1)



from . import Check, Finding
from .utils import is_index, get_first_heading


class TopLevelHeadingFinding(Finding):
    NAME = "TopLevelHeading"

    def msg(self):
        return "Document has no heading"


class TopLevelHeadingCheck(Check):
    FINDING = TopLevelHeadingFinding

    def run(self, file, lines):
        if not is_index(file):
            return
        _, i = get_first_heading(lines)
        if i == -1:
            self.add_finding(file, 1)



from .config.main import config_setup
from .checks.main import check_setup
from .version import __version__


def setup(app):
    config_setup(app)
    check_setup(app)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



class Finding:
    NAME = "Finding"

    def __init__(self, file, line, **kwargs):
        self.file = file
        self.line = line

    def msg(self):
        return "<Missing description>"

    def excerpt_lines(self, n):
        lines = []
        try:
            with open(self.file, "r", encoding="utf-8") as rst_file:
                all_lines = rst_file.readlines()

            if self.line > len(all_lines):
                raise IndexError

            start = max(0, self.line - n - 1)
            end = min(len(all_lines), self.line + n)

            for i in range(start, end):
                lines.append((i + 1, all_lines[i]))

        except FileNotFoundError:
            lines = [(self.line, self.file)]
        except IndexError:
            lines = [(self.line, self.file)]
        return lines

    def as_str(self, show_excerpt=0):
        display = f"[{self.NAME}] {self.msg()}\n"
        if show_excerpt > 0:
            for i, line in self.excerpt_lines(show_excerpt):
                if i == self.line:
                    display += f"{i:3d} -> {line}"
                else:
                    display += f"{i:3d}:  {line}"
        return display



from . import Check, Finding
from .utils import is_heading


class UnderlineLengthFinding(Finding):
    NAME = "UnderlineLength"

    def msg(self):
        return f"Heading underline length does not match title length"


class UnderlineLengthCheck(Check):
    FINDING = UnderlineLengthFinding

    def run(self, file, lines):
        prev_line_length = -1

        for i, line in enumerate(lines):
            line_length = len(line)
            if is_heading(line) and prev_line_length != line_length and prev_line_length != -1:
                self.add_finding(file, i + 1)

            prev_line_length = line_length



from pprint import pprint
from .checks.main import generate_options

if __name__ == "__main__":
    print("Available stylecheck options (and their defaults):\n")
    pprint(generate_options(), sort_dicts=False, width=100, indent=4)



import re
from typing import Union

heading_regex = re.compile(r"([=\-`:'\"~^_*+#<>])\1*$")
index_regex = re.compile(r".*/([^/]+)/doc/index.rst")


def is_heading(line: str) -> bool:
    return heading_regex.match(line) is not None


def is_index(file: str) -> bool:
    return index_regex.match(file) is not None


def get_module(file: str) -> Union[str, None]:
    match = index_regex.match(file)
    if match:
        return match.group(1)
    else:
        return None


def get_first_heading(lines):
    heading_candidate = None
    for i, line in enumerate(lines):
        if (
            is_heading(line)
            and heading_candidate is not None
            and len(line) >= len(heading_candidate)
        ):
            return heading_candidate
        heading_candidate = line, i
    return "", -1



import collections
from copy import deepcopy
from fnmatch import fnmatch
from sphinx.util import logging

from .top_level_casing_check import TopLevelCasingCheck
from .top_level_heading_check import TopLevelHeadingCheck
from .top_level_length_check import TopLevelLengthCheck
from .top_level_modulename_check import TopLevelModuleNameCheck
from .trailing_whitespace_check import TrailingWhitespacesCheck
from .heading_levels_check import HeadingLevelsCheck
from .underline_length_check import UnderlineLengthCheck
from .include_rst_check import IncludeRstCheck

logger = logging.getLogger(__name__)


all_checks = {
    "trailing_whitespaces": TrailingWhitespacesCheck,
    "heading_levels": HeadingLevelsCheck,
    "underline_length": UnderlineLengthCheck,
    "top_level_casing": TopLevelCasingCheck,
    "top_level_heading": TopLevelHeadingCheck,
    "top_level_length": TopLevelLengthCheck,
    "top_level_modulename": TopLevelModuleNameCheck,
    "include_rst": IncludeRstCheck,
}


global_options = {
    "exclude": [],
    "excerpt": 2,
}


common_check_options = {
    "enabled": True,
    "exclude": [],
}


def generate_options():
    options = global_options
    for key, check in all_checks.items():
        options[key] = deepcopy(check.CONFIG)
        options[key].update(common_check_options)

    return options


def updated_options_from_config(config):
    options = generate_options()

    def update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    update(options, config)

    return options


def style_check(_app, env, docnames):
    options = updated_options_from_config(env.config.stylecheck)

    enabled_checks = []
    for key, check in all_checks.items():
        if options[key]["enabled"]:
            check.CONFIG = deepcopy(options[key])
            enabled_checks.append(check())

    if not enabled_checks:
        logger.warning("No stylechecks enabled")

    for filepath, docname in [(f"{env.srcdir}/{docname}.rst", docname) for docname in docnames]:
        if any([fnmatch(docname, exclusion) for exclusion in options["exclude"]]):
            continue

        try:
            with open(filepath, "r", encoding="utf-8") as file:
                lines = file.read().splitlines()
        except FileNotFoundError:
            logger.error(f"Cannot open {filepath}")
            return

        for check in enabled_checks:
            if any([fnmatch(docname, exclusion) for exclusion in check.CONFIG["exclude"]]):
                continue
            check.run(filepath, lines)

    num_findings = 0
    for check in enabled_checks:
        for finding in check.findings:
            num_findings += 1
            excerpt = env.config.stylecheck.get("excerpt", global_options["excerpt"])
            logger.info(
                f"\n{finding.as_str(show_excerpt=excerpt)}",
                location=(finding.file, finding.line),
                color="red",
            )

    if num_findings > 0:
        logger.warning(f"Found {num_findings} style violations")


def check_setup(app):
    app.connect("env-before-read-docs", style_check)
    app.add_config_value("stylecheck", generate_options(), "env")



from . import Check, Finding
import re


class TrailingWhitespacesFinding(Finding):
    NAME = "TrailingWhitespace"

    def msg(self):
        return f"Line contains trailing whitespaces"


class TrailingWhitespacesCheck(Check):
    FINDING = TrailingWhitespacesFinding

    trailing_whitespace_regex = re.compile(r".*\s+$")

    def run(self, file, lines):
        for i, line in enumerate(lines):
            if self.trailing_whitespace_regex.match(line):
                self.add_finding(file, i + 1)
